/*
 * This file is part of the Disco Deterministic Network Calculator v2.2.0 "Hydra".
 *
 * Copyright (C) 2011 - 2015 Steffen Bondorf
 *
 * disco | Distributed Computer Systems Lab
 * University of Kaiserslautern, Germany
 *
 * http://disco.cs.uni-kl.de
 *
 *
 * The Disco Deterministic Network Calculator (DiscoDNC) is free software;
 * you can redistribute it and/or modify it under the terms of the 
 * GNU Lesser General Public License as published by the Free Software Foundation; 
 * either version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 *
 */

package unikl.disco.demos;

import java.util.Iterator;
import java.util.LinkedList;

import unikl.disco.curves.MaxServiceCurve;
import unikl.disco.curves.ServiceCurve;
import unikl.disco.curves.ArrivalCurve;
import unikl.disco.nc.PmooAnalysis;
import unikl.disco.nc.SeparateFlowAnalysis;
import unikl.disco.nc.TotalFlowAnalysis;
import unikl.disco.network.Link;
import unikl.disco.network.Flow;
import unikl.disco.network.Network;
import unikl.disco.network.Server;

/**
 * 
 * @author Steffen Bondorf
 */
public class Demo4
{
	public Demo4() {}

	public void run()
	{
		ServiceCurve service_curve = ServiceCurve.createRateLatency( 10.0e6, 0.01 );
		MaxServiceCurve max_service_curve = MaxServiceCurve.createRateLatency( 100.0e6, 0.001 );
		
		Network network = new Network();	

		try {
			int numServers = 9;
			Server[] servers = new Server[numServers];
			
			for ( int i = 0; i < numServers; i++ )
			{
				servers[i] = network.addServer( service_curve, max_service_curve );
				servers[i].setUseGamma( false );
				servers[i].setUseExtraGamma( false );
			}
			
			network.addLink( servers[0], servers[2] );
			network.addLink( servers[1], servers[2] );
			Link l_1_3 = network.addLink( servers[1], servers[3] );
			Link l_2_4 = network.addLink( servers[2], servers[4] );
			Link l_3_4 = network.addLink( servers[3], servers[4] );
			Link l_4_5 = network.addLink( servers[4], servers[5] );
			Link l_5_6 = network.addLink( servers[5], servers[6] );
			Link l_6_7 = network.addLink( servers[6], servers[7] );
			Link l_7_8 = network.addLink( servers[7], servers[8] );
			
			ArrivalCurve arrival_curve = ArrivalCurve.createTokenBucket( 0.1e6, 0.1 * 0.1e6 );
			
			LinkedList<Link> path0 = new LinkedList<Link>();
			
//			Links need to be ordered from source server to sink server when defining a path manually
			path0.add(l_2_4);
			path0.add(l_4_5);
			path0.add(l_5_6);
			path0.add(l_6_7);
			path0.add(l_7_8);
	
			network.addFlow( arrival_curve, path0 );
				
			LinkedList<Link> path1 = new LinkedList<Link>();
			path1.add(l_1_3);
			path1.add(l_3_4);
			path1.add(l_4_5);
			path1.add(l_5_6);
			
			network.addFlow( arrival_curve, path1 );
			
			Iterator<Flow> flow_iter = network.getFlows().iterator();
			while ( flow_iter.hasNext() ) {
				Flow flow_of_interest = flow_iter.next();
			
				System.out.println( "Flow of interest : " + flow_of_interest.toString() );
				System.out.println();

//				Analyze the network	
//				TFA
				System.out.println( "--- Total Flow Analysis ---" );
				TotalFlowAnalysis tfa = new TotalFlowAnalysis( network );
				
				try {
					tfa.performAnalysis( flow_of_interest );
					System.out.println( "delay bound     : " + tfa.getDelayBound() );
					System.out.println( "     per server : " + tfa.getServerDelayBoundMapString() );
					System.out.println( "backlog bound   : " + tfa.getBacklogBound() );
					System.out.println( "     per server : " + tfa.getServerBacklogBoundMapString() );
					System.out.println( "alpha per server: " + tfa.getServerAlphasMapString() );
				} catch (Exception e) {
					System.out.println( "TFA analysis failed" );
					System.out.println( e.toString() );
				}
				
				System.out.println();
		
//				SFA
				System.out.println( "--- Separated Flow Analysis ---" );
				SeparateFlowAnalysis sfa = new SeparateFlowAnalysis( network );
		
				try {
					sfa.performAnalysis( flow_of_interest );
					System.out.println( "e2e SFA SCs     : " + sfa.getLeftOverServiceCurves() );
					System.out.println( "     per server : " + sfa.getServerLeftOverBetasMapString() );
					System.out.println( "xtx per server  : " + sfa.getServerAlphasMapString() );
					System.out.println( "delay bound     : " + sfa.getDelayBound() );
					System.out.println( "backlog bound   : " + sfa.getBacklogBound() );
				} catch (Exception e) {
					System.out.println( "SFA analysis failed" );
					System.out.println( e.toString() );
				}
				
				System.out.println();
		
//				PMOO
				System.out.println( "--- PMOO Analysis ---" );
				PmooAnalysis pmoo = new PmooAnalysis( network );
				
				try {
					pmoo.performAnalysis( flow_of_interest );
					System.out.println( "e2e PMOO SCs    : " + pmoo.getLeftOverServiceCurves() );
					System.out.println( "xtx per server  : " + pmoo.getServerAlphasMapString() );
					System.out.println( "delay bound     : " + pmoo.getDelayBound() );
					System.out.println( "backlog bound   : " + pmoo.getBacklogBound() );
				} catch (Exception e) {
					System.out.println( "PMOO SFA analysis failed" );
					System.out.println( e.toString() );
				}

				System.out.println();
				System.out.println();
			}
		} catch (Exception e) {
			System.out.println( e.toString() );
		}
	}

	public static void main( String[] args )
	{
		Demo4 demo = new Demo4();
		demo.run();
	}
}
